# Copyright 2021 The ChromiumOS Authors
# Use of this source code is governed by a BSD-style license that can be
# found in the LICENSE file.

"""Tests for zmake version code."""

import datetime
import subprocess
from unittest import mock

# pylint: disable=import-error
import pytest
from zmake import version
import zmake.output_packers
import zmake.project


# pylint:disable=redefined-outer-name,unused-argument


def _git_init(repo):
    """Create a new git repository."""
    repo.mkdir()
    subprocess.run(
        ["git", "-c", "init.defaultBranch=main", "-C", repo, "init"], check=True
    )


def _git_add(repo, path, contents="example!\n"):
    """Write contents and stage a file."""
    path.write_text(contents)
    subprocess.run(["git", "-C", repo, "add", path], check=True)


def _git_commit(repo, message="message!"):
    env = {
        "GIT_AUTHOR_NAME": "Alyssa P. Hacker",
        "GIT_AUTHOR_EMAIL": "aphacker@example.org",
        "GIT_AUTHOR_DATE": "Thu, 07 Apr 2005 22:13:13 +0200",
        "GIT_COMMITTER_NAME": "Ben Bitdiddle",
        "GIT_COMMITTER_EMAIL": "bitdiddle@example.org",
        "GIT_COMMITTER_DATE": "Tue, 30 Aug 2005 10:50:30 -0700",
    }
    subprocess.run(
        ["git", "-C", repo, "commit", "-m", message], check=True, env=env
    )


def _setup_example_repos(tmp_path):
    """Setup temporary project and git repo

    Args:
        tmp_path: Directory to set up files in.

    Returns:
        A 3-tuple of project, project_version, git_path.
    """
    project_path = tmp_path / "prj"
    project_path.mkdir()

    project = zmake.project.Project(
        zmake.project.ProjectConfig(
            project_name="prj",
            zephyr_board="foo",
            output_packer=zmake.output_packers.RawBinPacker,
            supported_toolchains=["coreboot-sdk"],
            project_dir=project_path,
        ),
    )

    # Has two commits.
    git_path = tmp_path / "ec"
    _git_init(git_path)
    _git_add(git_path, git_path / "file1")
    _git_commit(git_path)
    _git_add(git_path, git_path / "file2")
    _git_commit(git_path)

    project_version = "123.456.789"

    return project, project_version, git_path


def test_version_string(fake_date, tmp_path):
    """Test developer build with version string provided."""
    project, ver, git_path = _setup_example_repos(tmp_path)
    assert (
        version.get_version_string(
            project.config.project_name, ver, git_path=git_path
        )
        == "prj-123.456.789-c0d52f1"
    )


def test_version_string_default(fake_date, tmp_path):
    """Test developer build without version string provided."""
    project, _, git_path = _setup_example_repos(tmp_path)
    assert (
        version.get_version_string(
            project.config.project_name,
            git_path=git_path,
        )
        == "prj-0.0.0-c0d52f1"
    )


def test_version_string_static(tmp_path):
    """Test static developer build with version string provided."""
    project, ver, git_path = _setup_example_repos(tmp_path)
    assert (
        version.get_version_string(
            project.config.project_name, ver, git_path=git_path, static=True
        )
        == "prj-123.456.789"
    )


def test_version_string_default_static(fake_date, tmp_path):
    """Test static developer build without version string provided."""
    project, _, git_path = _setup_example_repos(tmp_path)
    assert (
        version.get_version_string(
            project.config.project_name, git_path=git_path, static=True
        )
        == "prj-0.0.0"
    )


@pytest.fixture
def fake_user_hostname():
    """Fixture to provide a fake user and hostname."""
    with mock.patch("getpass.getuser", return_value="toukmond", autospec=True):
        with mock.patch("platform.node", return_value="pokey", autospec=True):
            yield


@pytest.fixture
def fake_date():
    """Fixture to provide a fake date."""
    fixed_date = datetime.datetime(2021, 6, 28, 3, 18, 53)
    with mock.patch("datetime.datetime") as mock_datetime:
        mock_datetime.now.return_value = fixed_date
        yield


HEADER_VERSION_STR = "prj-123.456.789-c0d52f1"
EXPECTED_HEADER = (
    "/* This file is automatically generated by zmake_tests */\n"
    '#define VERSION "prj-123.456.789-c0d52f1"\n'
    '#define CROS_EC_VERSION32 "prj-123.456.789-c0d52f1"\n'
    '#define BUILDER "toukmond@pokey"\n'
    '#define DATE "2021-06-28 03:18:53"\n'
)
HEADER_VERSION_STR_STATIC = "prj-123.456.789"
EXPECTED_HEADER_STATIC = (
    "/* This file is automatically generated by zmake_tests */\n"
    '#define VERSION "prj-123.456.789"\n'
    '#define CROS_EC_VERSION32 "prj-123.456.789"\n'
    '#define BUILDER "reproducible@build"\n'
    '#define DATE "STATIC_VERSION_DATE"\n'
)


def test_header_gen(fake_user_hostname, fake_date, tmp_path):
    """Test generating the version header."""
    # Test the simple case (static=False, no existing header).
    output_file = tmp_path / "ec_version.h"
    version.write_version_header(HEADER_VERSION_STR, output_file, "zmake_tests")
    assert output_file.read_text() == EXPECTED_HEADER


def test_header_gen_reproducible_build(tmp_path):
    """Test that reproducible builds produce the right header."""
    # With static=True this time.
    output_file = tmp_path / "ec_version.h"
    version.write_version_header(
        HEADER_VERSION_STR_STATIC, output_file, "zmake_tests", static=True
    )
    assert output_file.read_text() == EXPECTED_HEADER_STATIC


def test_header_gen_exists_not_changed(fake_user_hostname, fake_date, tmp_path):
    """Test that the version file is not changed."""
    # Test we don't overwrite if no changes needed.
    output_file = tmp_path / "ec_version.h"

    # First time, write and record mtime.
    version.write_version_header(HEADER_VERSION_STR, output_file, "zmake_tests")
    expected_mtime = output_file.stat().st_mtime

    # Do another write (contents should be unchanged).
    version.write_version_header(HEADER_VERSION_STR, output_file, "zmake_tests")

    # Assert we didn't write again.
    assert output_file.stat().st_mtime == expected_mtime


def test_header_gen_exists_needs_changes(
    fake_user_hostname, fake_date, tmp_path
):
    """Test that the version file is changed, when needed."""
    # Test we overwrite when it exists already and changes are needed.
    output_file = tmp_path / "ec_version.h"

    # First time, write and save contents.
    version.write_version_header(HEADER_VERSION_STR, output_file, "zmake_tests")
    original_contents = output_file.read_text()

    # Do another write (contents should be changed).
    version.write_version_header(
        HEADER_VERSION_STR_STATIC, output_file, "zmake_tests", static=True
    )

    # Assert we overwrote.
    assert output_file.read_text() != original_contents
